#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <termio.h>
#include <signal.h>
#include <security/pam_appl.h>

#define DEFAULT_USER "sbi"
#define DEFAULT_SERVICE "login"

char *gets_hidden(char *s, int n);
void setup(void);
void cleanup(int sig);
int conv(int num_msg,
    const struct pam_message **msg,
    struct pam_response **resp,
    void *appdata_ptr);
int authenticate(char *service, char *user, char *tty);
void usage(void);

struct termios save, noecho;

int main(int argc, char *argv[])
{
    extern char *optarg;
    extern int optind;
    int c;
    char *user = strdup(DEFAULT_USER);
    char *service = strdup(DEFAULT_SERVICE); /* XXX add error check */
    char *tty;
    
    setup();

    if ((tty = ttyname(0)) == NULL) {
        perror("ttyname");
        cleanup(1);
    }
    
    while ((c = getopt(argc, argv, "u:s:t:h")) != EOF) {
        switch (c) {
        case '?':
        case 'h':
            usage();
            cleanup(1);
            break;
        case 'u':
            free(user);
            user = strdup(optarg);
            break;
        case 't':
            free(tty);
            tty = strdup(optarg);
            break;
        case 's':
            free(service);
            service = strdup(optarg);
            break;  
        }
    }
    if (argc > optind) {
        usage();
        cleanup(1);
    }

    if (authenticate(service, user, tty))
        puts("Authentication succeeded.");
    else
        puts("Authentication failed.");

    free(user);
    free(service);

    cleanup(0);
}

void usage(void)
{
    printf("Usage: pam [-s service] [-u user] [-t tty]\n");
}

void setup(void)
{
    if (ioctl(0, TCGETS, &save) == -1) {
        perror("ioctl");
        exit(1);
    }
    memcpy(&noecho, &save, sizeof(struct termios));
    noecho.c_lflag &= ~ECHO;
    if (signal(SIGTERM, cleanup) == SIG_ERR ||
        signal(SIGINT, cleanup) == SIG_ERR) {
        perror("signal");
        exit(1);
    }
    
}

void cleanup(int sig)
{
    if (ioctl(0, TCSETS, &save) == -1) {
        perror("ioctl");
    }
    exit(sig);
}

char *gets_hidden(char *s, int n)
{
    char *i;

    if (ioctl(0, TCSETS, &noecho) == -1) {
        perror("ioctl");
        cleanup(1);
    }
    if (fgets(s, n, stdin) == NULL) {
        perror("fgets");
        cleanup(1);
    }
    for (i = s; *i != '\0' && *i != '\n'; i++)
        ;
    *i = '\0';
    if (ioctl(0, TCSETS, &save) == -1) {
        perror("ioctl");
        cleanup(1);
    }
    return s;
}

int conv(int num_msg,
    const struct pam_message **msg,
    struct pam_response **resp,
    void *appdata_ptr)
{
    struct pam_response *rep;
    int i;
    char buf[256];

    if ((rep = malloc(num_msg * sizeof (struct pam_response))) == NULL) {
        printf("Out of memory in conv()\n");
        cleanup(1);
    }

/* This is modeled after jwz's function in xscreensaver.  The sun man
 * pages are way too vague. 
 */
    for (i = 0; i < num_msg; i++) {
        char *c;

        puts(msg[i]->msg);
        switch(msg[i]->msg_style) {
        case PAM_PROMPT_ECHO_OFF:
            gets_hidden(buf, 256);
            rep[i].resp = strdup(buf);
            rep[i].resp_retcode = PAM_SUCCESS;
            break;
        case PAM_PROMPT_ECHO_ON:
            fgets(buf, 256, stdin);
            for (c = buf; *c != '\0' && *c != '\n'; c++)
                ;
            *c = '\0';
            rep[i].resp = strdup(buf);
            rep[i].resp_retcode = PAM_SUCCESS;
            break;
        case PAM_ERROR_MSG:
        case PAM_TEXT_INFO:
            rep[i].resp = NULL;
            rep[i].resp_retcode = PAM_SUCCESS;
            break;
        default:
            puts("Unexpected PAM message in conv()");
            exit(1);
        }
    }
    *resp = rep;
    return PAM_SUCCESS;
}

int authenticate(char *service, char *user, char *tty)
{
    pam_handle_t *pamh = NULL;
    struct pam_conv pc;
    int s;
    int error = 0;

    pc.conv = conv;
    pc.appdata_ptr = 0;
    
    if ((s = pam_start(service, user, &pc, &pamh)) != PAM_SUCCESS) {
        printf("pam_start(): %s\n", pam_strerror(NULL, s));
        ++error;
    }

    if ((s = pam_set_item(pamh, PAM_TTY, tty)) != PAM_SUCCESS) {
        printf("pam_set_item(): %s\n", pam_strerror(pamh, s));
        ++error;
    }

    if ((s = pam_authenticate(pamh, 0)) != PAM_SUCCESS) { 
        printf("pam_authenticate(): %s\n", pam_strerror(pamh, s));
        ++error;
    }

    if ((s = pam_end(pamh, 0)) != PAM_SUCCESS) {
        printf("pam_end(): %s\n", pam_strerror(pamh, s));
        ++error;
    }
    return (!error);
}

